
#include "postgres.h"
#include "fmgr.h"
#include "libpq/pqformat.h"
#include "access/gist.h"
#include "access/stratnum.h"
#include "utils/array.h"
#include "catalog/pg_type.h"
#include <limits.h>
#include <math.h>

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

#if INT_MAX < 0x7fffffff
#error Expected int type to be at least 32 bit wide
#endif
#if LLONG_MAX < 0x7fffffffffffffff
#error Expected long long type to be at least 64 bit wide
#endif

#define CFL_TIMESTAMP_MAX (((int64_t)1e15)-1)
#define CFL_TIMESTAMP_MIN (-CFL_TIMESTAMP_MAX)

typedef struct {
  int64_t timestamp;
  int32_t person;
} cfl_pstamp;

typedef struct {
  char header[VARHDRSZ];
  int32_t npersons;
  int64_t timestamp;
  int32_t persons[FLEXIBLE_ARRAY_MEMBER];
} cfl_pstamp_query;

typedef struct {
  int32_t lower_person;
  int32_t upper_person;
  int64_t lower_timestamp;
  int64_t upper_timestamp;
} cfl_pstamp_key;

#define CFL_HUGE_DISTANCE 1e38

static int cfl_compare_person(const int32_t *a, const int32_t *b) {
  if (*a < *b) return -1;
  if (*a > *b) return 1;
  return 0;
}

static void cfl_finalize_pstamp_query(
  cfl_pstamp_query *query, int npersons
) {
  int i, j;
  qsort(query->persons, npersons, sizeof(int32_t), (void *)cfl_compare_person);
  for (i=0, j=0; i<npersons; i++) {
    if (!i || query->persons[i] != query->persons[i-1]) {
      query->persons[j++] = query->persons[i];
    }
  }
  SET_VARSIZE(
    query,
    offsetof(cfl_pstamp_query, persons) + j * sizeof(int32_t)
  );
  query->npersons = j;
}

static int cfl_compare_pstamp_key_ptr(
  const cfl_pstamp_key **a, const cfl_pstamp_key **b
) {
  if ((*a)->lower_person < (*b)->lower_person) return -1;
  if ((*a)->lower_person > (*b)->lower_person) return 1;
  if ((*a)->lower_timestamp < (*b)->lower_timestamp) return -1;
  if ((*a)->lower_timestamp > (*b)->lower_timestamp) return 1;
  if ((*a) < (*b)) return -1;
  if ((*a) > (*b)) return 1;
  return 0;
}

static void cfl_pstamp_key_union(
  cfl_pstamp_key *dst, const cfl_pstamp_key *src
) {
  if (src->lower_person < dst->lower_person) {
    dst->lower_person = src->lower_person;
    dst->lower_timestamp = src->lower_timestamp;
  } else if (
    src->lower_person == dst->lower_person &&
    src->lower_timestamp < dst->lower_timestamp
  ) {
    dst->lower_timestamp = src->lower_timestamp;
  }
  if (src->upper_person > dst->upper_person) {
    dst->upper_person = src->upper_person;
    dst->upper_timestamp = src->upper_timestamp;
  } else if (
    src->upper_person == dst->upper_person &&
    src->upper_timestamp > dst->upper_timestamp
  ) {
    dst->upper_timestamp = src->upper_timestamp;
  }
}

PG_FUNCTION_INFO_V1(cfl_notimpl);
Datum cfl_notimpl(PG_FUNCTION_ARGS) {
  ereport(ERROR, (errmsg("not implemented by conflux extension")));
}

PG_FUNCTION_INFO_V1(cfl_pstamp_in);
Datum cfl_pstamp_in(PG_FUNCTION_ARGS) {
  char *str = PG_GETARG_CSTRING(0);
  int person;
  long long timestamp;
  int len;
  if (sscanf(str, "%d@%lld%n", &person, &timestamp, &len) >= 2) {
    if (
      !str[len] &&
      person >= INT32_MIN && person <= INT32_MAX &&
      timestamp >= INT64_MIN && timestamp <= INT64_MAX
    ) {
      cfl_pstamp *pstamp;
      if (timestamp < CFL_TIMESTAMP_MIN || timestamp > CFL_TIMESTAMP_MAX) {
        ereport(ERROR, (
          errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
          errmsg(
            "numeric value out of range for type pstamp: %lld",
            (long long)timestamp
          )
        ));
      }
      pstamp = (cfl_pstamp *)palloc(sizeof(cfl_pstamp));
      pstamp->person = person;
      pstamp->timestamp = timestamp;
      PG_RETURN_POINTER(pstamp);
    }
  }
  ereport(ERROR, (
    errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
    errmsg("invalid input syntax for type pstamp: \"%s\"", str)
  ));
}

PG_FUNCTION_INFO_V1(cfl_pstamp_out);
Datum cfl_pstamp_out(PG_FUNCTION_ARGS) {
  cfl_pstamp *pstamp = (cfl_pstamp *)PG_GETARG_POINTER(0);
  PG_RETURN_CSTRING(psprintf(
    "%d@%lld",
    (int)pstamp->person,
    (long long)pstamp->timestamp
  ));
}

PG_FUNCTION_INFO_V1(cfl_pstamp_query_in);
Datum cfl_pstamp_query_in(PG_FUNCTION_ARGS) {
  char *str = PG_GETARG_CSTRING(0);
  char *strptr;
  int npersons = 1, i;
  long long scanval;
  int len;
  cfl_pstamp_query *query;
  for (strptr=str; *strptr; strptr++) {
    if (*strptr == ',') {
      npersons++;
      if (npersons == INT_MAX || npersons == INT32_MAX) {
        goto cfl_pstamp_query_in_error;
      }
    }
  }
  if (
    npersons >
    (SIZE_MAX - offsetof(cfl_pstamp_query, persons)) / sizeof(int32_t)
  ) {
    goto cfl_pstamp_query_in_error;
  }
  query = (cfl_pstamp_query *)palloc(
    offsetof(cfl_pstamp_query, persons) + npersons * sizeof(int32_t)
  );
  strptr = str;
  if (*strptr == '@') npersons = 0;
  for (i=0; i<npersons; i++) {
    if (
      sscanf(strptr, i ? ",%lld%n" : "%lld%n", &scanval, &len) != 1 ||
      scanval < INT32_MIN || scanval > INT32_MAX
    ) {
      goto cfl_pstamp_query_in_error;
    }
    strptr += len;
    query->persons[i] = scanval;
  }
  if (
    sscanf(strptr, "@%lld%n", &scanval, &len) != 1 ||
    scanval < INT64_MIN || scanval > INT64_MAX
  ) {
    goto cfl_pstamp_query_in_error;
  }
  strptr += len;
  if (scanval > CFL_TIMESTAMP_MAX+1) scanval = CFL_TIMESTAMP_MAX+1;
  query->timestamp = scanval;
  if (*strptr) goto cfl_pstamp_query_in_error;
  cfl_finalize_pstamp_query(query, npersons);
  PG_RETURN_POINTER(query);
  cfl_pstamp_query_in_error:
  ereport(ERROR, (
    errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
    errmsg("invalid input syntax for type pstamp_query: \"%s\"", str)
  ));
}

PG_FUNCTION_INFO_V1(cfl_pstamp_query_out);
Datum cfl_pstamp_query_out(PG_FUNCTION_ARGS) {
  cfl_pstamp_query *query = (cfl_pstamp_query *)PG_DETOAST_DATUM(
    PG_GETARG_POINTER(0)
  );
  StringInfoData buf;
  int i;
  initStringInfo(&buf);
  for (i=0; i<query->npersons; i++) {
    appendStringInfo(&buf, i ? ",%d" : "%d", (int)query->persons[i]);
  }
  appendStringInfo(&buf, "@%lld", (long long)query->timestamp);
  PG_FREE_IF_COPY(query, 0);
  PG_RETURN_CSTRING(buf.data);
}

PG_FUNCTION_INFO_V1(cfl_pstamp_recv);
Datum cfl_pstamp_recv(PG_FUNCTION_ARGS) {
  StringInfo buf = (StringInfo)PG_GETARG_POINTER(0);
  cfl_pstamp *pstamp = (cfl_pstamp *)palloc(sizeof(cfl_pstamp));
  pstamp->person = pq_getmsgint(buf, 4);
  pstamp->timestamp = pq_getmsgint64(buf);
  PG_RETURN_POINTER(pstamp);
}

PG_FUNCTION_INFO_V1(cfl_pstamp_send);
Datum cfl_pstamp_send(PG_FUNCTION_ARGS) {
  cfl_pstamp *pstamp = (cfl_pstamp *)PG_GETARG_POINTER(0);
  StringInfoData buf;
  pq_begintypsend(&buf);
  pq_sendint(&buf, pstamp->person, 4);
  pq_sendint64(&buf, pstamp->timestamp);
  PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}

PG_FUNCTION_INFO_V1(cfl_pstamp_new);
Datum cfl_pstamp_new(PG_FUNCTION_ARGS) {
  int32_t person = PG_GETARG_INT32(0);
  int64_t timestamp = PG_GETARG_INT64(1);
  cfl_pstamp *pstamp;
  if (timestamp < CFL_TIMESTAMP_MIN || timestamp > CFL_TIMESTAMP_MAX) {
    ereport(ERROR, (
      errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
      errmsg(
        "numeric value out of range for type pstamp: %lld",
        (long long)timestamp
      )
    ));
  }
  pstamp = palloc(sizeof(cfl_pstamp));
  pstamp->person = person;
  pstamp->timestamp = timestamp;
  PG_RETURN_POINTER(pstamp);
}

PG_FUNCTION_INFO_V1(cfl_pstamp_query_new);
Datum cfl_pstamp_query_new(PG_FUNCTION_ARGS) {
  ArrayType *array;
  Datum *persons;
  bool *nulls;
  int npersons, i, j;
  cfl_pstamp_query *query;
  if (PG_ARGISNULL(0)) PG_RETURN_NULL();
  array = PG_GETARG_ARRAYTYPE_P(0);
  deconstruct_array(array, INT4OID, 4, true, 'i', &persons, &nulls, &npersons);
  if (
    npersons >= INT_MAX || npersons >= INT32_MAX ||
    npersons >
    (SIZE_MAX - offsetof(cfl_pstamp_query, persons)) / sizeof(int32_t)
  ) {
    ereport(ERROR, (
      errcode(ERRCODE_INVALID_PARAMETER_VALUE),
      errmsg("array passed to pstamp_query constructor is too long")
    ));
  }
  for (i=0, j=0; i<npersons; i++) { if (!nulls[i]) j++; }
  query = (cfl_pstamp_query *)palloc(
    offsetof(cfl_pstamp_query, persons) + j * sizeof(int32_t)
  );
  if (PG_ARGISNULL(1)) query->timestamp = CFL_TIMESTAMP_MAX+1;
  else {
    int64_t timestamp = PG_GETARG_INT64(1);
    if (timestamp > CFL_TIMESTAMP_MAX+1) timestamp = CFL_TIMESTAMP_MAX+1;
    query->timestamp = timestamp;
  }
  for (i=0, j=0; i<npersons; i++) {
    if (!nulls[i]) query->persons[j++] = DatumGetInt64(persons[i]);
  }
  pfree(persons);
  pfree(nulls);
  cfl_finalize_pstamp_query(query, j);
  PG_RETURN_POINTER(query);
}

PG_FUNCTION_INFO_V1(cfl_pstamp_match);
Datum cfl_pstamp_match(PG_FUNCTION_ARGS) {
  cfl_pstamp *pstamp = (cfl_pstamp *)PG_GETARG_POINTER(0);
  cfl_pstamp_query *query = (cfl_pstamp_query *)PG_DETOAST_DATUM(
    PG_GETARG_POINTER(1)
  );
  if (query->timestamp <= pstamp->timestamp) PG_RETURN_BOOL(false);
  if (bsearch(
    &pstamp->person,
    query->persons,
    query->npersons,
    sizeof(int32_t),
    (void *)cfl_compare_person
  )) {
    PG_FREE_IF_COPY(query, 1);
    PG_RETURN_BOOL(true);
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_BOOL(false);
}

PG_FUNCTION_INFO_V1(cfl_pstamp_distance);
Datum cfl_pstamp_distance(PG_FUNCTION_ARGS) {
  cfl_pstamp *pstamp = (cfl_pstamp *)PG_GETARG_POINTER(0);
  cfl_pstamp_query *query = (cfl_pstamp_query *)PG_DETOAST_DATUM(
    PG_GETARG_POINTER(1)
  );
  if (query->timestamp <= pstamp->timestamp) PG_RETURN_FLOAT8(INFINITY);
  if (bsearch(
    &pstamp->person,
    query->persons,
    query->npersons,
    sizeof(int32_t),
    (void *)cfl_compare_person
  )) {
    PG_FREE_IF_COPY(query, 1);
    PG_RETURN_FLOAT8(query->timestamp - pstamp->timestamp);
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_FLOAT8(INFINITY);
}

PG_FUNCTION_INFO_V1(cfl_gist_pstamp_consistent);
Datum cfl_gist_pstamp_consistent(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  cfl_pstamp_key *key = (cfl_pstamp_key *)DatumGetPointer(entry->key);
  cfl_pstamp_query *query = (cfl_pstamp_query *)PG_DETOAST_DATUM(
    PG_GETARG_POINTER(1)
  );
  bool *recheck = (bool *)PG_GETARG_POINTER(4);
  if (query->npersons) {
    int i = 0, j = query->npersons-1, k;
    while (i != j) {
      k = (i+j) / 2;
      if (query->persons[k] < key->lower_person) i = k + 1;
      else j = k;
    }
    if (
      query->persons[i] >= key->lower_person &&
      query->persons[i] <= key->upper_person
    ) {
      if (key->lower_person == key->upper_person) {
        if (key->lower_timestamp < query->timestamp) {
          *recheck = false;
          PG_FREE_IF_COPY(query, 1);
          PG_RETURN_BOOL(true);
        } else {
          PG_FREE_IF_COPY(query, 1);
          PG_RETURN_BOOL(false);
        }
      } else {
        *recheck = true;
        PG_FREE_IF_COPY(query, 1);
        PG_RETURN_BOOL(true);
      }
    }
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_BOOL(false);
}

PG_FUNCTION_INFO_V1(cfl_gist_pstamp_union);
Datum cfl_gist_pstamp_union(PG_FUNCTION_ARGS) {
  GistEntryVector *entryvec = (GistEntryVector *)PG_GETARG_POINTER(0);
  cfl_pstamp_key *key = palloc(sizeof(cfl_pstamp_key));
  int i;
  *key = *(cfl_pstamp_key *)DatumGetPointer(entryvec->vector[0].key);
  for (i=1; i<entryvec->n; i++) {
    cfl_pstamp_key_union(
      key,
      (cfl_pstamp_key *)DatumGetPointer(entryvec->vector[i].key)
    );
  }
  PG_RETURN_POINTER(key);
}

PG_FUNCTION_INFO_V1(cfl_gist_pstamp_compress);
Datum cfl_gist_pstamp_compress(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  GISTENTRY *retval;
  if (entry->leafkey) {
    cfl_pstamp *pstamp = (cfl_pstamp *)DatumGetPointer(entry->key);
    cfl_pstamp_key *key = palloc(sizeof(cfl_pstamp_key));
    key->lower_person = pstamp->person;
    key->upper_person = pstamp->person;
    key->lower_timestamp = pstamp->timestamp;
    key->upper_timestamp = pstamp->timestamp;
    retval = palloc(sizeof(GISTENTRY));
    gistentryinit(
      *retval, PointerGetDatum(key),
      entry->rel, entry->page, entry->offset, false
    );
  } else {
    retval = entry;
  }
  PG_RETURN_POINTER(retval);
}

PG_FUNCTION_INFO_V1(cfl_gist_decompress);
Datum cfl_gist_decompress(PG_FUNCTION_ARGS) {
  PG_RETURN_POINTER(PG_GETARG_POINTER(0));
}

PG_FUNCTION_INFO_V1(cfl_gist_pstamp_penalty);
Datum cfl_gist_pstamp_penalty(PG_FUNCTION_ARGS) {
  GISTENTRY *origentry = (GISTENTRY *)PG_GETARG_POINTER(0);
  GISTENTRY *newentry = (GISTENTRY *)PG_GETARG_POINTER(1);
  float *penalty = (float *)PG_GETARG_POINTER(2);
  cfl_pstamp_key *orig = (cfl_pstamp_key *)DatumGetPointer(origentry->key);
  cfl_pstamp_key grown = *orig;
  cfl_pstamp_key_union(
    &grown,
    (cfl_pstamp_key *)DatumGetPointer(newentry->key)
  );
  if (
    grown.lower_person == orig->lower_person &&
    grown.upper_person == orig->upper_person
  ) {
    if (
      grown.lower_timestamp == orig->lower_timestamp &&
      grown.upper_timestamp == orig->upper_timestamp
    ) {
      *penalty = 0.0;
    } else {
      *penalty = 1e-20 * (uint64_t)(
        (grown.upper_timestamp - grown.lower_timestamp) -
        (orig->upper_timestamp - orig->lower_timestamp)
      );
    }
  } else {
    *penalty = (uint32_t)(
      (grown.upper_person - grown.lower_person) -
      (orig->upper_person - orig->lower_person)
    );
  }
  PG_RETURN_POINTER(penalty);
}

PG_FUNCTION_INFO_V1(cfl_gist_pstamp_picksplit);
Datum cfl_gist_pstamp_picksplit(PG_FUNCTION_ARGS) {
  GistEntryVector *entryvec = (GistEntryVector *)PG_GETARG_POINTER(0);
  GIST_SPLITVEC *v = (GIST_SPLITVEC *)PG_GETARG_POINTER(1);
  OffsetNumber i;
  int nkeys = entryvec->n - FirstOffsetNumber;
  const cfl_pstamp_key **keys = palloc(nkeys * sizeof(cfl_pstamp_key *));
  const cfl_pstamp_key *key;
  const cfl_pstamp_key *middle;
  cfl_pstamp_key *unionL = NULL;
  cfl_pstamp_key *unionR = NULL;
  for (i=FirstOffsetNumber; i<entryvec->n; i=OffsetNumberNext(i)) {
    keys[i-FirstOffsetNumber] = (cfl_pstamp_key *)DatumGetPointer(entryvec->vector[i].key);
  }
  qsort(keys, nkeys, sizeof(cfl_pstamp_key *), (void *)cfl_compare_pstamp_key_ptr);
  middle = keys[nkeys/2];
  v->spl_left = (OffsetNumber *)palloc(entryvec->n * sizeof(OffsetNumber));
  v->spl_nleft = 0;
  v->spl_right = (OffsetNumber *)palloc(entryvec->n * sizeof(OffsetNumber));
  v->spl_nright = 0;
  for (i=FirstOffsetNumber; i<entryvec->n; i=OffsetNumberNext(i)) {
    key = (cfl_pstamp_key *)DatumGetPointer(entryvec->vector[i].key);
    if (cfl_compare_pstamp_key_ptr(&key, &middle) <= 0) {
      if (unionL) {
        cfl_pstamp_key_union(unionL, key);
      } else {
        unionL = palloc(sizeof(cfl_pstamp_key));
        *unionL = *key;
      }
      v->spl_left[v->spl_nleft++] = i;
    } else {
      if (unionR) {
        cfl_pstamp_key_union(unionR, key);
      } else {
        unionR = palloc(sizeof(cfl_pstamp_key));
        *unionR = *key;
      }
      v->spl_right[v->spl_nright++] = i;
    }
  }
  v->spl_ldatum = PointerGetDatum(unionL);
  v->spl_rdatum = PointerGetDatum(unionR);
  PG_RETURN_POINTER(v);
}

PG_FUNCTION_INFO_V1(cfl_gist_pstamp_same);
Datum cfl_gist_pstamp_same(PG_FUNCTION_ARGS) {
  cfl_pstamp_key *key1 = (cfl_pstamp_key *)PG_GETARG_POINTER(0);
  cfl_pstamp_key *key2 = (cfl_pstamp_key *)PG_GETARG_POINTER(1);
  bool *boolptr = (bool *)PG_GETARG_POINTER(2);
  *boolptr = (
    key1->lower_person == key2->lower_person &&
    key1->upper_person == key2->upper_person &&
    key1->lower_timestamp == key2->lower_timestamp &&
    key1->upper_timestamp == key2->upper_timestamp
  );
  PG_RETURN_POINTER(boolptr);
}

PG_FUNCTION_INFO_V1(cfl_gist_pstamp_distance);
Datum cfl_gist_pstamp_distance(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  cfl_pstamp_key *key = (cfl_pstamp_key *)DatumGetPointer(entry->key);
  cfl_pstamp_query *query = (cfl_pstamp_query *)PG_DETOAST_DATUM(
    PG_GETARG_POINTER(1)
  );
  if (query->npersons) {
    int i = 0, j = query->npersons-1, k;
    while (i != j) {
      k = (i+j) / 2;
      if (query->persons[k] < key->lower_person) i = k + 1;
      else j = k;
    }
    if (
      query->persons[i] >= key->lower_person &&
      query->persons[i] <= key->upper_person
    ) {
      if (key->lower_person == key->upper_person) {
        if (key->lower_timestamp < query->timestamp) {
          if (key->upper_timestamp >= query->timestamp) {
            PG_FREE_IF_COPY(query, 1);
            PG_RETURN_FLOAT8(0);
          } else {
            PG_FREE_IF_COPY(query, 1);
            PG_RETURN_FLOAT8(query->timestamp - key->upper_timestamp);
          }
        } else {
          PG_FREE_IF_COPY(query, 1);
          PG_RETURN_FLOAT8(CFL_HUGE_DISTANCE);
        }
      } else {
        PG_FREE_IF_COPY(query, 1);
        PG_RETURN_FLOAT8(0);
      }
    }
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_FLOAT8(CFL_HUGE_DISTANCE);
}

